﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Bridge;
using Bridge.Html5;
using Bridge.jQuery2;

namespace WebMrbc
{
	internal class EcnlTaskWorkspace : IClassWorkspace
	{
		BlocklyView view;
		internal Workspace workspace;
		private Ruby _RubyCode;

		public string Identifier { get { return view.Identifier; } }
		public Workspace Workspace { get { return workspace; } }
		public BlocklyView View { get { return view; } }
		public Ruby RubyCode { get { return _RubyCode; } }

		public EcnlTaskWorkspace(BlocklyView view)
		{
			this.view = view;
			workspace = view.Init();
		}

		public string GetImageUrl()
		{
			return "img/no_image.png";
		}

		public bool IsPreset()
		{
			return true;
		}

		public string ToCode(string filename)
		{
			if (Collections.LocalNode == null)
				return "";

			_RubyCode = new Ruby(filename);
			var enode = Collections.LocalNode.enode;
			var objects = new JsonObjectInfo[0];
			foreach (var item in Collections.ClassWorkspaces) {
				var obj = item as EObjectWorkspace;
				if ((obj == null) || (obj == Collections.LocalNode))
					continue;

				objects.Push(obj.eobject);
			}
			enode.objects = objects;
			var result = _RubyCode.defineEcnlTask(this, enode);
			view.Changed = false;
			return result;
		}

		//Element ele;

		public void Activate()
		{
			//ele = Blockly.Xml.workspaceToDom(workspace);
		}

		public void Inactivate()
		{
			//workspace.clear();
			//Blockly.Xml.domToWorkspace(ele, workspace);
		}

		public void ReloadToolbox(HTMLElement toolbox)
		{
			toolbox.AppendChild(Document.CreateElement("sep"));
			var xml = jQuery.ParseXML(App.ArduinoToolbox);
			var categories = xml.ChildNodes[0];
			foreach (var item in categories.ChildNodes) {
				if (item.NodeName != "category")
					continue;
				toolbox.AppendChild(item);
			}

			toolbox.AppendChild(Document.CreateElement("sep"));
			xml = jQuery.ParseXML(App.EcnlToolbox);
			categories = xml.ChildNodes[0];
			foreach (var item in categories.ChildNodes) {
				if (item.NodeName != "category")
					continue;
				toolbox.AppendChild(item);
			}

			var category = Document.CreateElement("category");
			category.SetAttribute("name", "ECNL SvcTask");
			category.SetAttribute("colour", "230");
			toolbox.AppendChild(category);

			var funcs = new[] { "setup", "recv_esv", "break_wait", "timeout", "snd_msg" };
			var procs = new[] { "is_local_addr", "is_multicast_addr", "is_valid_addrid", "get_local_addr", "get_multicast_addr", "get_remote_addr", "get_remote_id", "set_remote_addr", "add_remote_addr" };
			foreach (var name in funcs) {
				var block = Document.CreateElement("block");
				block.SetAttribute("type", ProceduresDefnoreturnBlock.type_name);

				var field = Document.CreateElement("field");
				field.SetAttribute("name", "NAME");
				field.AppendChild(Document.CreateTextNode(name));
				block.AppendChild(field);

				category.AppendChild(block);
			}

			foreach (var name in procs) {
				var block = Document.CreateElement("block");
				block.SetAttribute("type", ProceduresDefreturnBlock.type_name);

				var field = Document.CreateElement("field");
				field.SetAttribute("name", "NAME");
				field.AppendChild(Document.CreateTextNode(name));
				block.AppendChild(field);

				category.AppendChild(block);
			}
		}

		public void OpenModifyView(Action<bool> callback)
		{
			view.ReloadToolbox(this);
			callback(true);
		}

		public string Template(string template)
		{
			template = template.Replace("%identifier%", Identifier);
			template = template.Replace("%attribute%", "SvcTask");
			template = template.Replace("%img%", GetImageUrl());
			return template;
		}
	}

	partial class Ruby
	{
		internal string defineEcnlTask(EcnlTaskWorkspace workspace, JsonNodeInfo localNode)
		{
			global = false;
			string code = workspaceToCode(workspace.Workspace);
			global = true;

			var identifier = workspace.Identifier;
			var funcs = new string[0];

			var blocks = workspace.Workspace.getTopBlocks(true);
			for (var i = 0; i < blocks.Length; i++) {
				var block = blocks[i];
				string name;

				switch (block.type) {
				case ProceduresDefnoreturnBlock.type_name:
					name = block.getFieldValue("NAME");
					switch (name) {
					case "setup":
					case "recv_esv":
					case "break_wait":
					case "timeout":
					case "snd_msg":
						funcs.Push(name);
						break;
					}
					break;
				case ProceduresDefreturnBlock.type_name:
					name = block.getFieldValue("NAME");
					switch (name) {
					case "is_local_addr":
					case "is_multicast_addr":
					case "is_valid_addrid":
					case "get_local_addr":
					case "get_multicast_addr":
					case "get_remote_addr":
					case "get_remote_id":
					case "set_remote_addr":
					case "add_remote_addr":
						funcs.Push(name);
						break;
					}
					break;
				}
			}

			var sb = new StringBuilder();

			sb.AppendLine("class " + identifier + " < ECNL::SvcTask");
			sb.AppendLine("	def initialize()");
			sb.AppendLine("		@profile = " + localNode.identifier + ".new(" + localNode.instanceCode + ")");
			switch (localNode.objects.Length) {
			case 0:
				sb.AppendLine("		@devices = []");
				break;
			case 1: {
					var o = localNode.objects[0];
					sb.AppendLine("		@devices = [ " + o.identifier + ".new(" + o.instanceCode + ", @profile) ]");
				}
				break;
			default:
				sb.AppendLine("		@devices = [");
				foreach (var o in localNode.objects) {
					sb.AppendLine("			" + o.identifier + ".new(" + o.instanceCode + ", @profile)");
				}
				sb.AppendLine("		]");
				break;
			}
			sb.AppendLine("		@enodadrb_table = []");
			sb.AppendLine();
			sb.AppendLine("		super()");
			if (funcs.Contains("setup")) {
				sb.AppendLine();
				sb.AppendLine("		setup()");
			}
			sb.AppendLine("	end");
			sb.AppendLine();
			foreach (var line in code.Split("\n")) {
				sb.AppendLine("\t" + line);
			}
			sb.AppendLine();
			if (!funcs.Contains("recv_esv")) {
				sb.AppendLine("	def recv_esv(esv)");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("break_wait")) {
				sb.AppendLine("	def break_wait(data)");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("timeout")) {
				sb.AppendLine("	def timeout()");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("snd_msg")) {
				sb.AppendLine("	def snd_msg(ep, data)");
				sb.AppendLine("		# 通信レイヤーへ送信");
				sb.AppendLine("		TargetBoard::snd_msg(ep, data)");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("is_local_addr")) {
				sb.AppendLine("	def is_local_addr(ep)");
				sb.AppendLine("		TargetBoard::is_local_addr(ep)");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("is_multicast_addr")) {
				sb.AppendLine("	def is_multicast_addr(ep)");
				sb.AppendLine("		TargetBoard::is_multicast_addr(ep)");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("is_valid_addrid")) {
				sb.AppendLine("	def is_valid_addrid(id)");
				sb.AppendLine("		(id >= 0) && ((id - ECNL::ENOD_REMOTE_ID) < @enodadrb_table.length)");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("get_local_addr")) {
				sb.AppendLine("	def get_local_addr()");
				sb.AppendLine("		TargetBoard::get_local_addr()");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("get_multicast_addr")) {
				sb.AppendLine("	def get_multicast_addr()");
				sb.AppendLine("		TargetBoard::get_multicast_addr()");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("get_remote_addr")) {
				sb.AppendLine("	def get_remote_addr(id)");
				sb.AppendLine("		index = id - ECNL::ENOD_REMOTE_ID");
				sb.AppendLine("		if (index < 0) || (index >= @enodadrb_table.length)");
				sb.AppendLine("			nil");
				sb.AppendLine("		else");
				sb.AppendLine("			@enodadrb_table[index]");
				sb.AppendLine("		end");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("get_remote_id")) {
				sb.AppendLine("	# 通信レイヤーアドレスの同じものを検索");
				sb.AppendLine("	def get_remote_id(ep)");
				sb.AppendLine("		id = ECNL::ENOD_REMOTE_ID - 1");
				sb.AppendLine("		for ea in @enodadrb_table do");
				sb.AppendLine("			id += 1");
				sb.AppendLine("			if !ea then");
				sb.AppendLine("				next");
				sb.AppendLine("			end");
				sb.AppendLine("			if !TargetBoard::equals_addr(ea, ep) then");
				sb.AppendLine("				next");
				sb.AppendLine("			end");
				sb.AppendLine();
				sb.AppendLine("			return id");
				sb.AppendLine("		end");
				sb.AppendLine();
				sb.AppendLine("		return ECNL::ENOD_NOT_MATCH_ID");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("set_remote_addr")) {
				sb.AppendLine("	# 対応するリモートノードを検索");
				sb.AppendLine("	def set_remote_addr(edata, ep)");
				sb.AppendLine("		id = ECNL::ENOD_REMOTE_ID - 1");
				sb.AppendLine("		for ea in @enodadrb_table do");
				sb.AppendLine("			id += 1");
				sb.AppendLine("			if !ea then");
				sb.AppendLine("				next");
				sb.AppendLine("			end");
				sb.AppendLine("			if !is_match(svc, edata, ep) then");
				sb.AppendLine("				next");
				sb.AppendLine("			end");
				sb.AppendLine();
				sb.AppendLine("			# 対応するリモートノードがあれば通信レイヤーアドレスを設定");
				sb.AppendLine("			@enodadrb_table[id - ECNL::ENOD_REMOTE_ID] = ep");
				sb.AppendLine();
				sb.AppendLine("			return id");
				sb.AppendLine("		end");
				sb.AppendLine();
				sb.AppendLine("		return ECNL::ENOD_NOT_MATCH_ID");
				sb.AppendLine("	end");
			}
			sb.AppendLine();
			if (!funcs.Contains("add_remote_addr")) {
				sb.AppendLine("	# 空き領域を探して自動登録");
				sb.AppendLine("	def add_remote_addr(edata, ep)");
				sb.AppendLine("		id = ECNL::ENOD_REMOTE_ID - 1");
				sb.AppendLine("		for ea in @enodadrb_table do");
				sb.AppendLine("			id += 1");
				sb.AppendLine("			if ea then");
				sb.AppendLine("				next");
				sb.AppendLine("			end");
				sb.AppendLine();
				sb.AppendLine("			@enodadrb_table[id - ECNL::ENOD_REMOTE_ID] = ep");
				sb.AppendLine();
				sb.AppendLine("			return id");
				sb.AppendLine("		end");
				sb.AppendLine();
				sb.AppendLine("		if @enodadrb_table.length >= 100 then");
				sb.AppendLine("			return ECNL::ENOD_NOT_MATCH_ID");
				sb.AppendLine("		end");
				sb.AppendLine();
				sb.AppendLine("		id = @enodadrb_table.length;");
				sb.AppendLine("		@enodadrb_table[id] = ep");
				sb.AppendLine();
				sb.AppendLine("		return id + ECNL::ENOD_REMOTE_ID");
				sb.AppendLine("	end");
				sb.AppendLine("end");

				code = sb.ToString();
			}
			return code;
		}

		string itr_code()
		{
			var sb = new StringBuilder();

			sb.AppendLine("itr = esv.itr_ini()");
			sb.AppendLine("itr.itr_nxt()");
			sb.AppendLine("until itr.is_eof do");
			sb.AppendLine("	if itr.epc == 0xD6 then");
			sb.AppendLine("		# LEDをON");
			sb.AppendLine("		digitalWrite(2, 1)");
			sb.AppendLine("	end");
			sb.AppendLine("	itr.itr_nxt()");
			sb.AppendLine("end");

			return sb.ToString();
		}
	}
}
